import asyncio
import itertools
import logging
from typing import Any, AsyncIterable, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union
from urllib.parse import urljoin

from pyppeteer import launch
from pyppeteer.element_handle import ElementHandle
from pyppeteer.errors import PageError
from pyppeteer.network_manager import Request
from pyppeteer.page import Page

from ..base import ScraperAbstract
from ..rule import Selector, SelectorType, rule_grouper, rule_sorter
from ..scraped_data import ScrapedData

logger = logging.getLogger(__name__)


class PyppeteerScraper(ScraperAbstract):
    """
    Pyppeteer-based scraper
    """

    supports_sync = False

    def run(
        self,
        urls: Sequence[str],
        pages: int = 1,
        proxy: Optional[Dict] = None,
        output: Optional[str] = None,
        format: str = "json",
        follow_urls: bool = False,
        save_per_page: bool = False,
        ignore_robots_txt: bool = False,
        headless: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Executes Pyppeteer-based scraper.

        :param urls: List of website URLs.
        :param pages: Maximum number of pages to crawl before exiting (default=1). This is only used when a navigate handler is defined. # noqa
        :param proxy: Proxy settings.
        :param output: Output file. If not provided, prints in the terminal.
        :param format: Output file format. If not provided, uses the extension of the output file or defaults to json.
        :param follow_urls: Automatically follow URLs.
        :param save_per_page: Flag to save data on every page extraction or not. If not, saves all the data at the end.
        :param ignore_robots_txt: Flag to ignore robots.txt.

        :param headless: Enables headless browser. (default=True)
        """
        super(PyppeteerScraper, self).run(
            urls=urls,
            pages=pages,
            proxy=proxy,
            output=output,
            format=format,
            follow_urls=follow_urls,
            save_per_page=save_per_page,
            ignore_robots_txt=ignore_robots_txt,
            headless=headless,
            **kwargs,
        )

    def setup(self, page: Page = None) -> None:
        raise Exception("Sync is not supported.")  # pragma: no cover

    async def setup_async(self, page: Page = None) -> None:
        """
        Executes setup handlers

        :param page: Page.
        """
        assert page is not None

        await self.event_pre_setup_async(page)

        for rule in self.get_setup_rules(page.url):
            for element in await self._get_elements(page, rule.selector):
                await rule.handler(element, page)

        await self.event_post_setup_async(page)

    def navigate(self, page: Page = None) -> bool:
        raise Exception("Sync is not supported.")  # pragma: no cover

    async def navigate_async(self, page: Page = None) -> bool:
        """
        Executes navigate handlers

        :param page: Page.
        """
        assert page is not None
        for rule in self.get_navigate_rules(page.url):
            for element in await self._get_elements(page, rule.selector):
                await rule.handler(element, page)
                logger.info("Navigated to %s", page.url)
                return True
        return False

    async def _block_url_if_needed(self, request: Request) -> Any:
        url = request.url
        source_url = (
            request.headers.get("referer") or request.headers.get("origin") or request.headers.get("host") or url
        )
        if self.adblock.check_network_urls(
            url=url,
            source_url=source_url,
            request_type=request.resourceType,
        ):
            logger.info("URL %s has been blocked.", url)
            return await request.abort()
        else:
            return await request.continue_()

    def run_sync(
        self,
        pages: int,
        proxy: Optional[Dict],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        headless: bool = True,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError  # pragma: no cover

    async def run_async(
        self,
        pages: int,
        proxy: Optional[Dict],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        headless: bool = True,
        **kwargs: Any,
    ) -> None:
        launch_args: Dict[str, Any] = {"headless": headless, "args": ["--disable-notifications"]}
        if proxy:
            launch_args["args"] = [f"--proxy-server={proxy['server']}"]

        browser = await launch(options=launch_args)
        page = await browser.newPage()

        if proxy and proxy["username"] and proxy["password"]:
            await page.authenticate(credentials={"username": proxy["username"], "password": proxy["password"]})

        await page.setRequestInterception(True)
        page.on("request", lambda res: asyncio.ensure_future(self._block_url_if_needed(res)))

        for url in self.iter_urls():
            logger.info("Requesting url %s", url)
            try:
                await page.goto(url)
            except PageError as e:
                logger.warning(e)
                continue
            logger.info("Loaded page %s", page.url)
            if follow_urls:
                for element in await page.querySelectorAll("a"):
                    handle = await element.getProperty("href")
                    href = await handle.jsonValue()
                    if isinstance(href, str):
                        absolute = urljoin(page.url, href)
                        if absolute.rstrip("/") != page.url.rstrip("/"):
                            self.urls.append(absolute)

            await self.setup_async(page=page)

            for i in range(1, pages + 1):
                current_page = page.url
                self.collected_data.extend([data async for data in self.extract_all_async(page_number=i, page=page)])

                if save_per_page:
                    await self._save_async(format, output, save_per_page)

                if i == pages or not await self.navigate_async(page=page) or current_page == page.url:
                    break

        await browser.close()

    def collect_elements(self, page: Page = None) -> Iterable[Tuple[str, int, int, int, Any, Callable]]:
        raise Exception("Sync is not supported.")  # pragma: no cover

    async def extract_all_async(self, page_number: int, **kwargs: Any) -> AsyncIterable[ScrapedData]:
        """
        Extracts all the data using the registered handler functions.

        Modified version for Pyppeteer
        """
        page = kwargs["page"]

        collected_elements = [element async for element in self.collect_elements_async(**kwargs)]

        for page_url, group_index, group_id, element_index, element, handler in collected_elements:
            data = await handler(element, page)

            if not len(data):
                continue

            scraped_data = ScrapedData(
                page_number=page_number,
                page_url=page_url,
                group_id=group_id,
                group_index=group_index,
                element_index=element_index,
                data=data,
            )
            yield scraped_data

    async def collect_elements_async(
        self, page: Page = None
    ) -> AsyncIterable[Tuple[str, int, int, int, Any, Callable]]:
        """
        Collects all the elements and returns a generator of element-handler pair.
        """
        assert page is not None
        page_url = page.url
        for group_selector, g in itertools.groupby(
            sorted(self.get_scraping_rules(page_url), key=rule_sorter), key=rule_grouper
        ):
            rules = list(sorted(g, key=lambda r: r.priority))
            for group_index, group in enumerate(await self._get_elements(page, group_selector)):
                for rule in rules:
                    for element_index, element in enumerate(await self._get_elements(group, rule.selector)):
                        yield page_url, group_index, id(group), element_index, element, rule.handler

    @staticmethod
    async def _get_elements(parent: Union[ElementHandle, Page], selector: Selector) -> List[ElementHandle]:
        selector_str = selector.to_str()
        selector_type = selector.selector_type()
        if selector_type in (SelectorType.CSS, SelectorType.ANY):  # assume CSS
            return await parent.querySelectorAll(selector_str)
        elif selector_type == SelectorType.XPATH:
            return await parent.xpath(selector_str)
        elif selector_type == SelectorType.TEXT:
            escaped_selector = selector_str.replace('"', '\\"')
            return await parent.xpath(f".//*[contains(text(), '{escaped_selector}')]")
        else:
            raise Exception("Regex selector is not supported.")



import itertools
import logging
from typing import Any, AsyncIterable, Callable, Dict, Iterable, Optional, Sequence, Tuple, Union
from urllib.parse import urljoin

from playwright import async_api, sync_api
from playwright.async_api import async_playwright
from playwright.sync_api import sync_playwright

from .base import ScraperAbstract
from .rule import rule_grouper, rule_sorter

logger = logging.getLogger(__name__)


class PlaywrightScraper(ScraperAbstract):
    """
    Playwright-based scraper
    """

    def run(
        self,
        urls: Sequence[str],
        pages: int = 1,
        proxy: Optional[sync_api.ProxySettings] = None,
        output: Optional[str] = None,
        format: str = "json",
        follow_urls: bool = False,
        save_per_page: bool = False,
        ignore_robots_txt: bool = False,
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        """
        Executes Playwright-based scraper.

        :param urls: List of website URLs.
        :param pages: Maximum number of pages to crawl before exiting (default=1). This is only used when a navigate handler is defined. # noqa
        :param proxy: Proxy settings. (see https://playwright.dev/python/docs/api/class-apirequest#api-request-new-context-option-proxy)  # noqa
        :param output: Output file. If not provided, prints in the terminal.
        :param format: Output file format. If not provided, uses the extension of the output file or defaults to json.
        :param follow_urls: Automatically follow URLs.
        :param save_per_page: Flag to save data on every page extraction or not. If not, saves all the data at the end.
        :param ignore_robots_txt: Flag to ignore robots.txt.

        :param headless: Enables headless browser. (default=True)
        :param browser_type: Playwright supported browser types ("chromium", "webkit" or "firefox").
        """
        super(PlaywrightScraper, self).run(
            urls=urls,
            pages=pages,
            proxy=proxy,
            output=output,
            format=format,
            follow_urls=follow_urls,
            save_per_page=save_per_page,
            ignore_robots_txt=ignore_robots_txt,
            **{**kwargs, "headless": headless, "browser_type": browser_type},
        )

    @staticmethod
    def _query_selector_all(
        page_or_element: Union[sync_api.Page, sync_api.ElementHandle], selector: str
    ) -> Iterable[sync_api.ElementHandle]:
        """
        Temporary fix for coverage not counting page.query_selector_all() when used in a for loop.
        """
        yield from page_or_element.query_selector_all(selector)

    def setup(self, page: sync_api.Page = None) -> None:
        """
        Executes setup handlers

        :param page: Page.
        """
        assert page is not None

        self.event_pre_setup(page)

        for rule in self.get_setup_rules(page.url):
            for element in self._query_selector_all(page, rule.selector.to_str(with_type=True)):
                rule.handler(element, page)

        self.event_post_setup(page)

    async def setup_async(self, page: async_api.Page = None) -> None:
        """
        Executes setup handlers

        :param page: Page.
        """
        assert page is not None

        await self.event_pre_setup_async(page)

        for rule in self.get_setup_rules(page.url):
            for element in await page.query_selector_all(rule.selector.to_str(with_type=True)):
                await rule.handler(element, page)

        await self.event_post_setup_async(page)

    def navigate(self, page: sync_api.Page = None) -> bool:
        """
        Executes navigate handlers

        :param page: Page.
        """
        assert page is not None
        for rule in self.get_navigate_rules(page.url):
            for element in self._query_selector_all(page, rule.selector.to_str(with_type=True)):
                rule.handler(element, page)
                logger.info("Navigated to %s", page.url)
                return True
        return False

    async def navigate_async(self, page: async_api.Page = None) -> bool:
        """
        Executes navigate handlers

        :param page: Page.
        """
        assert page is not None
        for rule in self.get_navigate_rules(page.url):
            for element in await page.query_selector_all(rule.selector.to_str(with_type=True)):
                await rule.handler(element, page)
                logger.info("Navigated to %s", page.url)
                return True
        return False

    @staticmethod
    def _get_launch_kwargs(browser_type: str) -> Dict[str, Any]:
        args = []
        if browser_type == "chromium":
            args.append("--disable-notifications")
        return {"args": args, "firefox_user_prefs": {"dom.webnotifications.enabled": False}}

    def _block_url_if_needed(self, route: Union[sync_api.Route, async_api.Route]) -> Any:
        url = route.request.url
        source_url = (
            route.request.headers.get("referer")
            or route.request.headers.get("origin")
            or route.request.headers.get("host")
            or url
        )
        if self.adblock.check_network_urls(
            url=url,
            source_url=source_url,
            request_type=route.request.resource_type,
        ):
            logger.info("URL %s has been blocked.", url)
            return route.abort()
        return route.continue_()

    def run_sync(
        self,
        pages: int,
        proxy: Optional[sync_api.ProxySettings],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        launch_kwargs = self._get_launch_kwargs(browser_type)
        # FIXME: Coverage fails to cover anything within this context manager block
        with sync_playwright() as p:
            browser = p[browser_type].launch(headless=headless, proxy=proxy, **launch_kwargs)
            for url in self.iter_urls():
                page = browser.new_page()
                page.route("**/*", self._block_url_if_needed)
                logger.info("Requesting url %s", url)
                try:
                    page.goto(url)
                except sync_api.Error as e:
                    logger.warning(e)
                    page.close()
                    continue
                logger.info("Loaded page %s", page.url)
                if follow_urls:
                    for link in page.query_selector_all("a"):
                        absolute = urljoin(page.url, link.get_attribute("href"))
                        if absolute.rstrip("/") != page.url.rstrip("/"):
                            self.urls.append(absolute)

                self.setup(page=page)

                for i in range(1, pages + 1):
                    current_page = page.url
                    self.collected_data.extend(self.extract_all(page_number=i, page=page))

                    if save_per_page:
                        self._save(format, output, save_per_page)

                    if i == pages or not self.navigate(page=page) or current_page == page.url:
                        page.close()
                        break

                page.close()

            browser.close()

    async def run_async(
        self,
        pages: int,
        proxy: Optional[sync_api.ProxySettings],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        launch_kwargs = self._get_launch_kwargs(browser_type)
        async with async_playwright() as p:
            browser = await p[browser_type].launch(headless=headless, proxy=proxy, **launch_kwargs)
            for url in self.iter_urls():
                page = await browser.new_page()
                await page.route("**/*", self._block_url_if_needed)
                logger.info("Requesting url %s", url)
                try:
                    await page.goto(url)
                except async_api.Error as e:
                    logger.warning(e)
                    await page.close()
                    continue
                logger.info("Loaded page %s", page.url)
                if follow_urls:
                    for link in await page.query_selector_all("a"):
                        absolute = urljoin(page.url, await link.get_attribute("href"))
                        if absolute.rstrip("/") != page.url.rstrip("/"):
                            self.urls.append(absolute)

                await self.setup_async(page=page)

                for i in range(1, pages + 1):
                    current_page = page.url
                    self.collected_data.extend(
                        [data async for data in self.extract_all_async(page_number=i, page=page)]
                    )

                    if save_per_page:
                        await self._save_async(format, output, save_per_page)

                    if i == pages or not await self.navigate_async(page=page) or current_page == page.url:
                        await page.close()
                        break

                await page.close()

            await browser.close()

    def collect_elements(self, page: sync_api.Page = None) -> Iterable[Tuple[str, int, int, int, Any, Callable]]:
        """
        Collects all the elements and returns a generator of element-handler pair.
        """
        assert page is not None
        page_url = page.url
        for group_selector, g in itertools.groupby(
            sorted(self.get_scraping_rules(page_url), key=rule_sorter), key=rule_grouper
        ):
            rules = list(sorted(g, key=lambda r: r.priority))

            for group_index, group in enumerate(page.query_selector_all(group_selector.to_str(with_type=True))):
                for rule in rules:
                    for element_index, element in enumerate(
                        self._query_selector_all(group, rule.selector.to_str(with_type=True))
                    ):
                        yield page_url, group_index, id(group), element_index, element, rule.handler

    async def collect_elements_async(
        self, page: async_api.Page = None
    ) -> AsyncIterable[Tuple[str, int, int, int, Any, Callable]]:
        """
        Collects all the elements and returns a generator of element-handler pair.
        """
        assert page is not None
        page_url = page.url
        for group_selector, g in itertools.groupby(
            sorted(self.get_scraping_rules(page_url), key=rule_sorter), key=rule_grouper
        ):
            rules = list(sorted(g, key=lambda r: r.priority))
            for group_index, group in enumerate(await page.query_selector_all(group_selector.to_str(with_type=True))):
                for rule in rules:
                    for element_index, element in enumerate(
                        await group.query_selector_all(rule.selector.to_str(with_type=True))
                    ):
                        yield page_url, group_index, id(group), element_index, element, rule.handler



import itertools
import logging
from typing import Any, AsyncIterable, Callable, Iterable, Optional, Sequence, Tuple
from urllib.parse import urljoin

import httpx
from httpx._types import ProxiesTypes
from parsel import Selector as ParselSelector

from ..base import ScraperAbstract
from ..rule import Selector, SelectorType, rule_grouper, rule_sorter
from .utils import HTTPXMixin, async_http_get, http_get

logger = logging.getLogger(__name__)


class ParselScraper(ScraperAbstract, HTTPXMixin):
    """
    Scraper using Parsel parser backend and HTTPX for requests
    """

    def run(
        self,
        urls: Sequence[str],
        pages: int = 1,
        proxy: ProxiesTypes = None,
        output: Optional[str] = None,
        format: str = "json",
        follow_urls: bool = False,
        save_per_page: bool = False,
        ignore_robots_txt: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Executes Parsel-based scraper.

        :param urls: List of website URLs.
        :param pages: Maximum number of pages to crawl before exiting (default=1). This is only used when a navigate handler is defined. # noqa
        :param proxy: Proxy settings. (see https://www.python-httpx.org/advanced/#http-proxying)  # noqa
        :param output: Output file. If not provided, prints in the terminal.
        :param format: Output file format. If not provided, uses the extension of the output file or defaults to json.
        :param follow_urls: Automatically follow URLs.
        :param save_per_page: Flag to save data on every page extraction or not. If not, saves all the data at the end.
        :param ignore_robots_txt: Flag to ignore robots.txt.
        """
        super(ParselScraper, self).run(
            urls=urls,
            pages=pages,
            proxy=proxy,
            output=output,
            format=format,
            follow_urls=follow_urls,
            save_per_page=save_per_page,
            ignore_robots_txt=ignore_robots_txt,
            **kwargs,
        )

    def run_sync(
        self,
        pages: int,
        proxy: Optional[ProxiesTypes],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        **kwargs: Any,
    ) -> None:
        with httpx.Client(proxies=proxy, event_hooks={"request": [self._block_httpx_request_if_needed]}) as client:
            for request in self.iter_requests():
                logger.info("Requesting url %s - %s", request.method, request.url)
                for i in range(1, pages + 1):
                    content, url = http_get(client, request)
                    if not content:
                        break

                    selector = ParselSelector(content, base_url=url)
                    if follow_urls:
                        for link in selector.root.iterlinks():
                            absolute = urljoin(url, link[2])
                            if absolute.rstrip("/") != url.rstrip("/"):
                                self.urls.append(absolute)

                    self.setup(selector)

                    self.collected_data.extend(self.extract_all(page_number=i, selector=selector, url=url))

                    if save_per_page:
                        self._save(format, output, save_per_page)

                    if i == pages or not self.navigate():
                        break

    async def run_async(
        self,
        pages: int,
        proxy: Optional[ProxiesTypes],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        **kwargs: Any,
    ) -> None:
        async with httpx.AsyncClient(
            proxies=proxy, event_hooks={"request": [self._async_block_httpx_request_if_needed]}
        ) as client:
            for request in self.iter_requests():
                logger.info("Requesting url %s - %s", request.method, request.url)
                for i in range(1, pages + 1):
                    content, url = await async_http_get(client, request)
                    if not content:
                        break

                    selector = ParselSelector(content, base_url=url)
                    if follow_urls:
                        for link in selector.root.iterlinks():
                            absolute = urljoin(url, link[2])
                            if absolute.rstrip("/") != url.rstrip("/"):
                                self.urls.append(absolute)

                    await self.setup_async(selector)

                    self.collected_data.extend(
                        [data async for data in self.extract_all_async(page_number=i, selector=selector, url=url)]
                    )

                    if save_per_page:
                        await self._save_async(format, output, save_per_page)

                    if i == pages or not await self.navigate_async():
                        break

    def setup(self, selector: ParselSelector = None) -> None:
        """
        This will only call the pre-setup and post-setup events if extra actions are needed to the selector object.
        :param selector: Selector object
        """
        assert selector is not None
        self.event_pre_setup(selector)
        self.event_post_setup(selector)

    async def setup_async(self, selector: ParselSelector = None) -> None:
        """
        This will only call the pre-setup and post-setup events if extra actions are needed to the selector object.
        :param selector: Selector object
        """
        assert selector is not None
        await self.event_pre_setup_async(selector)
        await self.event_post_setup_async(selector)

    def navigate(self) -> bool:
        return False

    async def navigate_async(self) -> bool:
        return False

    def collect_elements(
        self, selector: ParselSelector = None, url: str = None
    ) -> Iterable[Tuple[str, int, int, int, Any, Callable]]:
        assert selector is not None
        assert url is not None

        for group_selector, g in itertools.groupby(
            sorted(self.get_scraping_rules(url), key=rule_sorter), key=rule_grouper
        ):
            rules = list(sorted(g, key=lambda r: r.priority))

            for group_index, group in enumerate(self._get_elements(selector, group_selector)):
                for rule in rules:
                    for element_index, element in enumerate(self._get_elements(group, rule.selector)):
                        yield url, group_index, id(group), element_index, element, rule.handler

    @staticmethod
    def _get_elements(parsel_selector: ParselSelector, selector: Selector) -> Iterable[ParselSelector]:
        selector_str = selector.to_str()
        selector_type = selector.selector_type()
        if selector_type in (SelectorType.CSS, SelectorType.ANY):  # assume CSS
            yield from parsel_selector.css(selector_str)
        elif selector_type == SelectorType.XPATH:
            yield from parsel_selector.xpath(selector_str)
        elif selector_type == SelectorType.TEXT:
            yield from parsel_selector.xpath(".//*[contains(text(), $string)]/text()", string=selector_str)
        elif selector_type == SelectorType.REGEX:
            yield from parsel_selector.re(selector_str)

    async def collect_elements_async(self, **kwargs: Any) -> AsyncIterable[Tuple[str, int, int, int, Any, Callable]]:
        for item in self.collect_elements(**kwargs):
            yield item



import logging
import sys
from typing import Dict, List, Optional, Set

logger = logging.getLogger(__name__)


def save_json(data: List[Dict], output: Optional[str]) -> bool:
    """
    Saves data to JSON.

    :param data: List of scraped data.
    :param output: Optional filename. If not provided, prints the data to stdout.
    :return: Success
    """
    if output is not None:
        _save_json(data, output)
    else:
        import json

        json.dump(data, sys.stdout, indent=2)
    return True


def _save_json(data: List[Dict], output: str) -> None:  # pragma: no cover
    import json

    with open(output, "w") as f:
        json.dump(data, f, indent=2)
    logger.info("%d items saved to %s.", len(data), output)


def save_csv(data: List[Dict], output: Optional[str]) -> bool:
    """
    Saves data to CSV.

    :param data: List of scraped data.
    :param output: Optional filename. If not provided, prints the data to stdout.
    :return: Success
    """
    if output is not None:
        _save_csv(data, output)
    else:
        import json

        # TODO: find a better way to present a table if output is not None
        logger.warning("Printing CSV to terminal is currently not supported. Defaulting to json.")
        json.dump(data, sys.stdout, indent=2)
    return True


def save_yaml(data: List[Dict], output: Optional[str]) -> bool:
    """
    Saves data to YAML.

    :param data: List of scraped data.
    :param output: Optional filename. If not provided, prints the data to stdout.
    :return: Success
    """

    if output is not None:
        _save_yaml(data, output)
    else:
        import yaml

        yaml.safe_dump(data, sys.stdout)
    return True


def _save_csv(data: List[Dict], output: str) -> None:  # pragma: no cover
    import csv

    headers: Set[str] = set()
    rows = []
    for item in data:
        headers.update(item.keys())
        rows.append(item)
    with open(output, "w") as f:
        writer = csv.DictWriter(f, fieldnames=sorted(headers))
        writer.writeheader()
        writer.writerows(rows)
    logger.info("%d items saved to %s.", len(data), output)


def _save_yaml(data: List[Dict], output: str) -> None:  # pragma: no cover
    import yaml

    with open(output, "w") as f:
        yaml.safe_dump(data, f)
    logger.info("%d items saved to %s.", len(data), output)



import asyncio
import collections
import itertools
import logging
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    Any,
    AsyncIterable,
    Callable,
    Coroutine,
    DefaultDict,
    Deque,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from urllib.error import URLError
from urllib.parse import urljoin, urlparse
from urllib.robotparser import RobotFileParser

from braveblock import Adblocker

from .rule import Rule, Selector, rule_filter
from .scraped_data import ScrapedData, scraped_data_grouper, scraped_data_sorter
from .storage import save_csv, save_json, save_yaml

logger = logging.getLogger(__name__)


class ScraperBase(ABC):
    """
    Base Scraper class.

    This registers all the selector and saving rules.
    """

    supports_sync = True
    current_url = ""

    def __init__(
        self,
        rules: List[Rule] = None,
        groups: Dict[Callable, Selector] = None,
        save_rules: Dict[Tuple[str, bool], Any] = None,
        events: Optional[DefaultDict] = None,
        has_async: bool = False,
        requests: Optional[Deque] = None,  # only valid for BeautifulSoup4, lxml and Parsel backends
        scraper: Optional["ScraperAbstract"] = None,
    ) -> None:
        self.rules: List[Rule] = rules or []
        self.groups: Dict[Callable, Selector] = groups or {}
        if save_rules is None:
            save_rules = {}
        self.save_rules: Dict[Tuple[str, bool], Any] = {
            ("json", False): save_json,
            ("csv", False): save_csv,
            ("yml", False): save_yaml,
            ("yaml", False): save_yaml,
            **save_rules,
        }
        self.events: DefaultDict = events or collections.defaultdict(list)
        self.has_async = has_async
        self.scraper = scraper
        self.adblock = Adblocker()
        self.urls: Deque = collections.deque()  # allows dynamically appending new URLs for crawling
        self.requests: Deque = requests or collections.deque()  # allows dynamically appending new requests for crawling
        self.allowed_domains: Set[str] = set()
        self.ignore_robots_txt: bool = False

    @abstractmethod
    def run(
        self,
        urls: Sequence[str],
        pages: int,
        proxy: Optional[Any],
        output: Optional[str],
        format: str,
        follow_urls: bool = False,
        save_per_page: bool = False,
        ignore_robots_txt: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Abstract method for executing the scraper.

        :param urls: List of website URLs.
        :param pages: Maximum number of pages to crawl before exiting (default=1). This is only used when a navigate handler is defined. # noqa
        :param proxy: Proxy settings.
        :param output: Output file. If not provided, prints in the terminal.
        :param format: Output file format. If not provided, uses the extension of the output file or defaults to json.
        :param follow_urls: Automatically follow URLs.
        :param save_per_page: Flag to save data on every page extraction or not. If not, saves all the data at the end.
        :param ignore_robots_txt: Flag to ignore robots.txt.
        """
        self.initialize_scraper(urls)
        self.ignore_robots_txt = ignore_robots_txt

        logger.info("Using %s...", self.__class__.__name__)

        if self.has_async or not self.supports_sync:
            logger.info("Using async mode...")
            loop = asyncio.get_event_loop()
            # FIXME: Tests fail on Python 3.7 when using asyncio.run()
            loop.run_until_complete(
                self.run_async(  # type: ignore
                    pages=pages,
                    proxy=proxy,
                    output=output,
                    format=format,
                    follow_urls=follow_urls,
                    save_per_page=save_per_page,
                    **kwargs,
                )
            )
            if not save_per_page:
                loop.run_until_complete(self._save_async(format, output, save_per_page))  # type: ignore
        else:
            logger.info("Using sync mode...")
            self.run_sync(  # type: ignore
                pages=pages,
                proxy=proxy,
                output=output,
                format=format,
                follow_urls=follow_urls,
                save_per_page=save_per_page,
                **kwargs,
            )
            if not save_per_page:
                self._save(format, output, save_per_page)  # type: ignore

        self.event_shutdown()

    def select(
        self,
        selector: str = None,
        group: str = None,
        setup: bool = False,
        navigate: bool = False,
        url_match: Union[str, Callable] = "*",
        priority: int = 100,
        css: str = None,
        xpath: str = None,
        text: str = None,
        regex: str = None,
        group_css: str = None,
        group_xpath: str = None,
        group_text: str = None,
        group_regex: str = None,
    ) -> Callable:
        """
        Decorator to register a handler function to a given selector.

        :param selector: Element selector (CSS, XPath, text, regex).
        :param group: (Optional) Element selector where the matched element should be grouped. Defaults to ":root".
        :param setup: Flag to register a setup handler.
        :param navigate: Flag to register a navigate handler.
        :param url_match: URL pattern matcher. Run the handler function only when the pattern matches (defaults to *) or when custom function/lambda returns True. # noqa
        :param priority: Priority, the lowest value will be executed first (default 100).
        :param css: CSS selector.
        :param xpath: XPath selector.
        :param text: Text selector.
        :param regex: Regular expression selector
        :param group_css: Group CSS selector.
        :param group_xpath: Group XPath selector.
        :param group_text: Group Text selector.
        :param group_regex: Group Regular expression selector
        """

        def wrapper(func: Callable) -> Union[Callable, Coroutine]:
            sel = Selector(selector=selector, css=css, xpath=xpath, text=text, regex=regex)
            assert sel, "Any of selector, css, xpath, text and regex params should be present."

            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            rule = Rule(
                selector=sel,
                group=Selector(selector=group, css=group_css, xpath=group_xpath, text=group_text, regex=group_regex),
                url_matcher=url_match,
                handler=func,
                setup=setup,
                navigate=navigate,
                priority=priority,
            )
            rules = self.scraper.rules if self.scraper else self.rules
            rules.append(rule)
            return func

        return wrapper

    def group(
        self,
        selector: str = None,
        css: str = None,
        xpath: str = None,
        text: str = None,
        regex: str = None,
    ) -> Callable:
        """
        Decorator to register a handler function to a given group.

        :param selector: Element selector (any of CSS, XPath, text, regex).
        :param css: CSS selector.
        :param xpath: XPath selector.
        :param text: Text selector.
        :param regex: Regular expression selector
        """

        def wrapper(func: Callable) -> Union[Callable, Coroutine]:
            if not (selector or css or xpath or text or regex):
                raise Exception("Any of selector, css, xpath, text or regex selectors must be present")

            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            group = Selector(selector=selector, css=css, xpath=xpath, text=text, regex=regex)
            if not self.scraper:
                if func in self.groups:
                    logger.warning(
                        "Group '%s' already exists for function '%s'. Skipping '%s'...",
                        self.groups[func],
                        func.__name__,
                        group,
                    )
                else:
                    self.groups[func] = group
            else:
                if func in self.scraper.groups:
                    logger.warning(
                        "Group '%s' already exists for function '%s'. Skipping '%s'...",
                        self.scraper.groups[func],
                        func.__name__,
                        group,
                    )
                else:
                    self.scraper.groups[func] = group
            return func

        return wrapper

    def save(self, format: str, is_per_page: bool = False) -> Callable:
        """
        Decorator to register a save function to a format.

        :param format: Format (json, csv, or any custom string).
        :param is_per_page: Flag to identify if func will be called after each page.
        """

        def wrapper(func: Callable) -> Callable:
            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            save_rules = self.scraper.save_rules if self.scraper else self.save_rules
            save_rules[format, is_per_page] = func
            return func

        return wrapper

    def startup(self) -> Callable:
        """
        Decorator to register a function to startup events.

        Startup events are executed before any actual scraping happens to, for example, setup databases, etc.
        """

        def wrapper(func: Callable) -> Callable:
            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            events = self.scraper.events if self.scraper else self.events
            events["startup"].append(func)
            return func

        return wrapper

    def pre_setup(self) -> Callable:
        """
        Decorator to register a function to pre-setup events.

        Pre-setup events are executed after a page is loaded (or HTTP response in case of HTTPX) and
        before running the setup functions.
        """

        def wrapper(func: Callable) -> Callable:
            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            events = self.scraper.events if self.scraper else self.events
            events["pre-setup"].append(func)
            return func

        return wrapper

    def post_setup(self) -> Callable:
        """
        Decorator to register a function to post-setup events.

        Post-setup events are executed after running the setup functions and before the actual web-scraping happens.
        This is useful when "page clean-ups" are done in the setup functions.
        """

        def wrapper(func: Callable) -> Callable:
            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            events = self.scraper.events if self.scraper else self.events
            events["post-setup"].append(func)
            return func

        return wrapper

    def shutdown(self) -> Callable:
        """
        Decorator to register a function to the shutdown events.

        Shutdown events are executed before terminating the application for cleaning up or closing resources like
        files and database sessions.
        """

        def wrapper(func: Callable) -> Callable:
            if asyncio.iscoroutinefunction(func):
                self.has_async = True

            events = self.scraper.events if self.scraper else self.events
            events["shutdown"].append(func)
            return func

        return wrapper

    def start_requests(self) -> Callable:
        """
        Decorator to register custom Request objects.
        """
        from httpx import Request

        def wrapper(func: Callable) -> Callable:
            requests = self.scraper.requests if self.scraper else self.requests
            for request in func():
                assert isinstance(request, Request)
                requests.append(request)
            return func

        return wrapper

    def get_current_url(self) -> str:
        return self.scraper.current_url if self.scraper else self.current_url

    def follow_url(self, url: str) -> None:
        self.scraper.urls.append(url) if self.scraper else self.urls.append(url)

    def iter_urls(self) -> Iterator[str]:
        try:
            while True:
                url = self.urls.popleft()
                if urlparse(url).netloc not in self.allowed_domains:
                    logger.info("URL %s is not in allowed domains.", url)
                    continue
                can_fetch, crawl_delay = self.can_fetch_and_crawl_delay(url)
                if not can_fetch:
                    logger.info("Not allowed to crawl %s", url)
                    continue
                time.sleep(crawl_delay)
                self.current_url = url
                yield url
        except IndexError:
            pass

    def can_fetch_and_crawl_delay(self, url: str) -> Tuple[bool, int]:
        if self.ignore_robots_txt:
            return True, 0
        user_agent = "dude"  # TODO: https://github.com/roniemartinez/dude/issues/63
        # TODO: Store content so as to prevent re-fetching
        robots_url = urljoin(url, "/robots.txt")
        parser = RobotFileParser(url=robots_url)
        try:
            parser.read()
        except URLError:
            parser.parse([""])
        crawl_delay = parser.crawl_delay(user_agent) or 0
        can_fetch = parser.can_fetch(user_agent, url)
        return can_fetch, int(crawl_delay)

    def _update_rule_groups(self) -> Iterable[Rule]:
        for rule in self.rules:
            if rule.group:
                yield rule
            elif rule.handler in self.groups:
                yield rule._replace(group=self.groups[rule.handler])
            else:
                yield rule._replace(group=Selector(selector=":root"))

    def initialize_scraper(self, urls: Sequence[str]) -> None:
        self.rules = [rule for rule in self._update_rule_groups()]
        self.urls = collections.deque(urls)
        self.allowed_domains = {urlparse(url).netloc for url in urls}
        self.event_startup()

    def event_startup(self) -> None:
        """
        Run all startup events
        """
        self.run_event("startup")

    def event_shutdown(self) -> None:
        """
        Run all shutdown events
        """
        self.run_event("shutdown")

    def run_event(self, event_name: str) -> None:
        loop = None
        if self.has_async:
            loop = asyncio.get_event_loop()

        for func in self.events[event_name]:
            if asyncio.iscoroutinefunction(func):
                assert loop is not None
                loop.run_until_complete(func())
            else:
                func()


class ScraperAbstract(ScraperBase):
    def __init__(
        self,
        rules: List[Rule] = None,
        groups: Dict[Callable, Selector] = None,
        save_rules: Dict[Tuple[str, bool], Any] = None,
        events: Optional[DefaultDict] = None,
        has_async: bool = False,
        requests: Optional[Deque] = None,
    ) -> None:
        super(ScraperAbstract, self).__init__(rules, groups, save_rules, events, has_async, requests)
        self.collected_data: List[ScrapedData] = []

    @abstractmethod
    async def run_async(
        self,
        pages: int,
        proxy: Optional[Any],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def run_sync(
        self,
        pages: int,
        proxy: Optional[Any],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def setup(self) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    async def setup_async(self) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def navigate(self) -> bool:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    async def navigate_async(self) -> bool:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def collect_elements(self) -> Iterable[Tuple[str, int, int, int, Any, Callable]]:
        """
        Collects all the elements and returns a generator of element-handler pair.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    async def collect_elements_async(self) -> AsyncIterable[Tuple[str, int, int, int, Any, Callable]]:
        """
        Collects all the elements and returns a generator of element-handler pair.
        """

        yield "", 0, 0, 0, 0, str  # HACK: mypy does not identify this as AsyncIterable  # pragma: no cover
        raise NotImplementedError  # pragma: no cover

    def event_pre_setup(self, *args: Any) -> None:
        """
        Run all pre-setup events.
        """
        for func in self.events["pre-setup"]:
            func(*args)

    def event_post_setup(self, *args: Any) -> None:
        """
        Run all post-setup events.
        """
        for func in self.events["post-setup"]:
            func(*args)

    async def event_pre_setup_async(self, *args: Any) -> None:
        """
        Run all pre-setup events.
        """
        for func in self.events["pre-setup"]:
            await func(*args)

    async def event_post_setup_async(self, *args: Any) -> None:
        """
        Run all post-setup events.
        """
        for func in self.events["post-setup"]:
            await func(*args)

    def extract_all(self, page_number: int, **kwargs: Any) -> Iterable[ScrapedData]:
        """
        Extracts all the data using the registered handler functions.
        """
        collected_elements = list(self.collect_elements(**kwargs))

        for page_url, group_index, group_id, element_index, element, handler in collected_elements:
            data = handler(element)

            if not data:
                continue

            scraped_data = ScrapedData(
                page_number=page_number,
                page_url=page_url,
                group_id=group_id,
                group_index=group_index,
                element_index=element_index,
                data=data,
            )
            yield scraped_data

    async def extract_all_async(self, page_number: int, **kwargs: Any) -> AsyncIterable[ScrapedData]:
        """
        Extracts all the data using the registered handler functions.
        """

        collected_elements = [element async for element in self.collect_elements_async(**kwargs)]

        for page_url, group_index, group_id, element_index, element, handler in collected_elements:
            data = await handler(element)

            if not data:
                continue

            scraped_data = ScrapedData(
                page_number=page_number,
                page_url=page_url,
                group_id=group_id,
                group_index=group_index,
                element_index=element_index,
                data=data,
            )
            yield scraped_data

    def get_scraping_rules(self, url: str) -> Iterable[Rule]:
        return filter(rule_filter(url), self.rules)

    def get_setup_rules(self, url: str) -> Iterable[Rule]:
        return sorted(filter(rule_filter(url, setup=True), self.rules), key=lambda r: r.priority)

    def get_navigate_rules(self, url: str) -> Iterable[Rule]:
        return sorted(filter(rule_filter(url, navigate=True), self.rules), key=lambda r: r.priority)

    def get_flattened_data(self) -> List[Dict]:
        items = []
        for _, g in itertools.groupby(sorted(self.collected_data, key=scraped_data_sorter), key=scraped_data_grouper):
            item: Dict = {}
            for d in g:
                for k, v in d._asdict().items():
                    if k == "data":
                        # FIXME: Keys defined in handler functions might duplicate predefined keys
                        item.update(**v)
                    elif k not in item:
                        item[f"_{k}"] = v
            items.append(item)
        return items

    def _save(self, format: str, output: Optional[str] = None, save_per_page: bool = False) -> None:
        if output:
            extension = Path(output).suffix.lower()[1:]
            format = extension
        try:
            handler = self.save_rules[format, save_per_page]
            data = self.get_flattened_data()
            if not len(data):
                logger.info(
                    "No data was scraped. Skipped saving %s.",
                    dict(format=format, output=format, save_per_page=save_per_page),
                )
                return
            if handler(data, output):
                self.collected_data.clear()
            else:
                raise Exception("Failed to save output %s.", {"output": output, "format": format})
        except KeyError:
            raise

    async def _save_async(self, format: str, output: Optional[str] = None, save_per_page: bool = False) -> None:
        if output:
            extension = Path(output).suffix.lower()[1:]
            format = extension
        try:
            handler = self.save_rules[format, save_per_page]
            data = self.get_flattened_data()
            if not len(data):
                logger.info(
                    "No data was scraped. Skipped saving %s.",
                    dict(format=format, output=format, save_per_page=save_per_page),
                )
                return
            if asyncio.iscoroutinefunction(handler):
                is_successful = await handler(data, output)
            else:
                is_successful = handler(data, output)
            if is_successful:
                self.collected_data.clear()
            else:
                raise Exception("Failed to save output %s.", {"output": output, "format": format})
        except KeyError:
            raise



import logging

from .scraper import Scraper

logger = logging.getLogger(__name__)

_scraper = Scraper()
group = _scraper.group
run = _scraper.run
save = _scraper.save
select = _scraper.select
startup = _scraper.startup
shutdown = _scraper.shutdown
pre_setup = _scraper.pre_setup
post_setup = _scraper.post_setup
start_requests = _scraper.start_requests
get_current_url = _scraper.get_current_url
follow_url = _scraper.follow_url



import asyncio
import itertools
import logging
from typing import Any, AsyncIterable, Callable, Iterable, Optional, Sequence, Tuple, Union
from urllib.parse import urljoin

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from seleniumwire.request import Request
from seleniumwire.webdriver import Chrome, Firefox
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.core.utils import ChromeType
from webdriver_manager.firefox import GeckoDriverManager

from ..base import ScraperAbstract
from ..rule import Selector, SelectorType, rule_grouper, rule_sorter
from .utils import get_chromedriver_latest_release

logger = logging.getLogger(__name__)


class SeleniumScraper(ScraperAbstract):
    """
    Selenium-based scraper
    """

    def run(
        self,
        urls: Sequence[str],
        pages: int = 1,
        proxy: Optional[Any] = None,
        output: Optional[str] = None,
        format: str = "json",
        follow_urls: bool = False,
        save_per_page: bool = False,
        ignore_robots_txt: bool = False,
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        """
        Executes Selenium-based scraper.

        :param urls: List of website URLs.
        :param pages: Maximum number of pages to crawl before exiting (default=1). This is only used when a navigate handler is defined. # noqa
        :param proxy: Proxy settings.
        :param output: Output file. If not provided, prints in the terminal.
        :param format: Output file format. If not provided, uses the extension of the output file or defaults to json.
        :param follow_urls: Automatically follow URLs.
        :param save_per_page: Flag to save data on every page extraction or not. If not, saves all the data at the end.
        :param ignore_robots_txt: Flag to ignore robots.txt.

        :param headless: Enables headless browser. (default=True)
        :param browser_type: Selenium supported browser types ("chromium", "firefox").
        """
        super(SeleniumScraper, self).run(
            urls=urls,
            pages=pages,
            proxy=proxy,
            output=output,
            format=format,
            follow_urls=follow_urls,
            save_per_page=save_per_page,
            ignore_robots_txt=ignore_robots_txt,
            headless=headless,
            browser_type=browser_type,
            **kwargs,
        )

    def setup(self, driver: WebDriver = None) -> None:
        """
        Executes setup handlers

        :param driver: Web Driver.
        """
        assert driver is not None

        self.event_pre_setup(driver)

        for rule in self.get_setup_rules(driver.current_url):
            for element in self._get_elements(driver, rule.selector):
                rule.handler(element, driver)

        self.event_post_setup(driver)

    async def setup_async(self, driver: WebDriver = None) -> None:
        """
        Executes setup handlers

        :param driver: Web Driver.
        """
        assert driver is not None

        await self.event_pre_setup_async(driver)

        for rule in self.get_setup_rules(driver.current_url):
            for element in self._get_elements(driver, rule.selector):
                if asyncio.iscoroutinefunction(rule.handler):
                    await rule.handler(element, driver)
                else:
                    rule.handler(element, driver)

        await self.event_post_setup_async(driver)

    def navigate(self, driver: WebDriver = None) -> bool:
        """
        Executes navigate handlers

        :param driver: WebDriver.
        """
        assert driver is not None
        for rule in self.get_navigate_rules(driver.current_url):
            for element in self._get_elements(driver, rule.selector):
                rule.handler(element, driver)
                logger.info("Navigated to %s", driver.current_url)
                return True
        return False

    async def navigate_async(self, driver: WebDriver = None) -> bool:
        """
        Executes navigate handlers

        :param driver: WebDriver.
        """
        assert driver is not None
        for rule in self.get_navigate_rules(driver.current_url):
            for element in self._get_elements(driver, rule.selector):
                if asyncio.iscoroutinefunction(rule.handler):
                    await rule.handler(element, driver)
                else:
                    rule.handler(element, driver)
                logger.info("Navigated to %s", driver.current_url)
                return True
        return False

    def run_sync(
        self,
        pages: int,
        proxy: Optional[Any],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        driver = self._get_driver(browser_type, headless)

        for url in self.iter_urls():
            logger.info("Requesting url %s", url)
            try:
                driver.get(url)
            except WebDriverException as e:
                logger.warning(e)
                continue
            logger.info("Loaded page %s", driver.current_url)
            if follow_urls:
                for link in driver.find_elements(by=By.CSS_SELECTOR, value="a"):
                    absolute = urljoin(driver.current_url, link.get_attribute("href"))
                    if absolute.rstrip("/") != driver.current_url.rstrip("/"):
                        self.urls.append(absolute)

            self.setup(driver=driver)

            for i in range(1, pages + 1):
                current_page = driver.current_url
                self.collected_data.extend(self.extract_all(page_number=i, driver=driver))

                if save_per_page:
                    self._save(format, output, save_per_page)

                if i == pages or not self.navigate(driver=driver) or current_page == driver.current_url:
                    break

        driver.quit()

    async def run_async(
        self,
        pages: int,
        proxy: Optional[Any],
        output: Optional[str],
        format: str,
        follow_urls: bool,
        save_per_page: bool,
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        driver = self._get_driver(browser_type, headless)

        for url in self.iter_urls():
            logger.info("Requesting url %s", url)
            try:
                driver.get(url)
            except WebDriverException as e:
                logger.warning(e)
                continue
            logger.info("Loaded page %s", driver.current_url)
            if follow_urls:
                for link in driver.find_elements(by=By.CSS_SELECTOR, value="a"):
                    absolute = urljoin(driver.current_url, link.get_attribute("href"))
                    if absolute.rstrip("/") != driver.current_url.rstrip("/"):
                        self.urls.append(absolute)

            await self.setup_async(driver=driver)

            for i in range(1, pages + 1):
                current_page = driver.current_url
                self.collected_data.extend(
                    [data async for data in self.extract_all_async(page_number=i, driver=driver)]
                )

                if save_per_page:
                    await self._save_async(format, output, save_per_page)

                if i == pages or not await self.navigate_async(driver=driver) or current_page == driver.current_url:
                    break

        driver.quit()
        await self._save_async(format, output, save_per_page)

    def _block_url_if_needed(self, request: Request) -> None:
        url = request.url
        source_url = (
            request.headers.get("referer") or request.headers.get("origin") or request.headers.get("host") or url
        )
        if self.adblock.check_network_urls(
            url=url,
            source_url=source_url,
            request_type=request.headers.get("sec-fetch-dest") or "other",
        ):
            logger.info("URL %s has been blocked.", url)
            request.abort()

    def _get_driver(self, browser_type: str, headless: bool) -> WebDriver:
        # TODO: Add more drivers: https://github.com/SergeyPirogov/webdriver_manager#webdriver-manager-for-python
        if browser_type == "firefox":
            executable_path = GeckoDriverManager().install()
            firefox_options = FirefoxOptions()
            firefox_options.headless = headless
            firefox_options.set_preference("dom.webnotifications.enabled", False)
            firefox_options.set_preference("network.captive-portal-service.enabled", False)
            driver = Firefox(service=FirefoxService(executable_path=executable_path), options=firefox_options)
        else:
            chrome_options = ChromeOptions()
            chrome_options.headless = headless
            chrome_options.add_argument("--no-sandbox")
            chrome_options.add_argument("--disable-notifications")
            executable_path = ChromeDriverManager(
                chrome_type=ChromeType.CHROMIUM, version=get_chromedriver_latest_release()
            ).install()
            driver = Chrome(service=ChromeService(executable_path=executable_path), options=chrome_options)

        driver.request_interceptor = self._block_url_if_needed

        return driver

    def collect_elements(self, driver: WebDriver = None) -> Iterable[Tuple[str, int, int, int, Any, Callable]]:
        """
        Collects all the elements and returns a generator of element-handler pair.
        """
        assert driver is not None
        page_url = driver.current_url
        for group_selector, g in itertools.groupby(
            sorted(self.get_scraping_rules(page_url), key=rule_sorter), key=rule_grouper
        ):
            rules = list(sorted(g, key=lambda r: r.priority))

            for group_index, group in enumerate(self._get_elements(driver, group_selector)):
                for rule in rules:
                    for element_index, element in enumerate(self._get_elements(group, rule.selector)):
                        yield page_url, group_index, id(group), element_index, element, rule.handler

    async def collect_elements_async(self, **kwargs: Any) -> AsyncIterable[Tuple[str, int, int, int, Any, Callable]]:
        for item in self.collect_elements(**kwargs):
            yield item

    @staticmethod
    def _get_elements(parent: Union[WebDriver, WebElement], selector: Selector) -> Iterable[WebElement]:
        selector_str = selector.to_str()
        selector_type = selector.selector_type()
        if selector_type in (SelectorType.CSS, SelectorType.ANY):  # assume CSS
            yield from parent.find_elements(by=By.CSS_SELECTOR, value=selector_str)
        elif selector_type == SelectorType.XPATH:
            yield from parent.find_elements(by=By.XPATH, value=selector_str)
        elif selector_type == SelectorType.TEXT:
            escaped_selector = selector_str.replace('"', '\\"')
            yield from parent.find_elements(by=By.XPATH, value=f".//*[contains(text(), '{escaped_selector}')]")
        elif selector_type == SelectorType.REGEX:
            raise Exception("Regex selector is not supported.")



import logging
from typing import Any, Optional, Sequence, Type

from .base import ScraperBase
from .playwright_scraper import PlaywrightScraper

logger = logging.getLogger(__name__)


class Scraper(ScraperBase):
    """
    Convenience class to easily use the available decorators.
    """

    def run(
        self,
        urls: Sequence[str],
        pages: int = 1,
        proxy: Optional[Any] = None,
        output: Optional[str] = None,
        format: str = "json",
        follow_urls: bool = False,
        save_per_page: bool = False,
        ignore_robots_txt: bool = False,
        # extra args
        parser: str = "playwright",
        headless: bool = True,
        browser_type: str = "chromium",
        **kwargs: Any,
    ) -> None:
        """
        Convenience method to handle switching between different types of parser backends.

        :param urls: List of website URLs.
        :param pages: Maximum number of pages to crawl before exiting (default=1). This is only used when a navigate handler is defined. # noqa
        :param proxy: Proxy settings.
        :param output: Output file. If not provided, prints in the terminal.
        :param format: Output file format. If not provided, uses the extension of the output file or defaults to json.
        :param follow_urls: Automatically follow URLs.
        :param save_per_page: Flag to save data on every page extraction or not. If not, saves all the data at the end.
        :param ignore_robots_txt: Flag to ignore robots.txt.

        :param parser: Parser backend ["playwright" (default), "bs4", "parsel, "lxml", "pyppeteer" or "selenium"]
        :param headless: Enables headless browser. (default=True)
        :param browser_type: Playwright supported browser types ("chromium", "chrome", "webkit", or "firefox").
        """

        logger.info("Scraper started...")

        if not self.scraper:
            scraper_class: Type[ScraperBase]
            if parser == "bs4":
                from .optional.beautifulsoup_scraper import BeautifulSoupScraper

                scraper_class = BeautifulSoupScraper
            elif parser == "parsel":
                from .optional.parsel_scraper import ParselScraper

                scraper_class = ParselScraper
            elif parser == "lxml":
                from .optional.lxml_scraper import LxmlScraper

                scraper_class = LxmlScraper
            elif parser == "pyppeteer":
                from .optional.pyppeteer_scraper import PyppeteerScraper

                scraper_class = PyppeteerScraper
            elif parser == "selenium":
                from .optional.selenium_scraper import SeleniumScraper

                scraper_class = SeleniumScraper
            else:
                scraper_class = PlaywrightScraper

            self.scraper = scraper_class(
                rules=self.rules,
                groups=self.groups,
                save_rules=self.save_rules,
                events=self.events,
                has_async=self.has_async,
                requests=self.requests,
            )

        if not ignore_robots_txt:
            logger.info(
                f"""robots.txt is currently not ignored.
        {"=" * 80}
        Any rules/restrictions set in a website's robots.txt, will be followed by default.
        To ignore robots.txt, add `--ignore-robots-txt` to CLI arguments or  pass `ignore_robots_txt=True` to `run()`
        {"=" * 80}""",
            )

        self.scraper.run(
            urls=urls,
            pages=pages,
            proxy=proxy,
            output=output,
            format=format,
            follow_urls=follow_urls,
            save_per_page=save_per_page or follow_urls,
            ignore_robots_txt=ignore_robots_txt,
            **{"headless": headless, "browser_type": browser_type},
        )



import fnmatch
from enum import Enum, auto
from typing import Callable, NamedTuple, Optional, Tuple, Union


class SelectorType(Enum):
    ANY = auto()
    CSS = auto()
    XPATH = auto()
    TEXT = auto()
    REGEX = auto()


class Selector(NamedTuple):
    selector: Optional[str] = None
    css: Optional[str] = None
    xpath: Optional[str] = None
    text: Optional[str] = None
    regex: Optional[str] = None

    def to_str(self, with_type: bool = False) -> str:
        if not with_type or self.selector:
            selector = self.selector or self.css or self.xpath or self.text or self.regex
            assert selector is not None
            return selector
        if self.css:
            return f"css={self.css}"
        elif self.xpath:
            return f"xpath={self.xpath}"
        elif self.text:
            return f"text={self.text}"
        return f"text=/{self.regex}/i"  # NOTE: Playwright support only

    def selector_type(self) -> SelectorType:
        if self.selector:
            return SelectorType.ANY
        elif self.css:
            return SelectorType.CSS
        elif self.xpath:
            return SelectorType.XPATH
        elif self.text:
            return SelectorType.TEXT
        elif self.regex:
            return SelectorType.REGEX
        raise Exception("No selector specified.")

    def __bool__(self) -> bool:
        return (self.selector or self.css or self.xpath or self.text or self.regex) is not None

    def __str__(self) -> str:
        return self.selector or self.css or self.xpath or self.text or self.regex or ""

    def __repr__(self) -> str:
        return self.to_str(with_type=True)  # pragma: no cover

    def __lt__(self, other: "Selector") -> bool:  # type: ignore
        return self.to_str(with_type=True) < other.to_str(with_type=True)


class Rule(NamedTuple):
    group: Selector
    selector: Selector
    url_matcher: Union[str, Callable]
    handler: Callable
    setup: bool
    navigate: bool
    priority: int


def rule_sorter(rule: Rule) -> Tuple[Selector, Selector]:
    return rule.group, rule.selector


def rule_grouper(rule: Rule) -> Selector:
    return rule.group


def rule_filter(url: str, setup: bool = False, navigate: bool = False) -> Callable:
    def wrapper(rule: Rule) -> bool:
        if callable(rule.url_matcher):
            matches = rule.url_matcher(url)
        else:
            matches = fnmatch.fnmatch(url, rule.url_matcher)
        return matches and rule.setup is setup and rule.navigate is navigate

    return wrapper



from typing import Dict, NamedTuple, Tuple


class ScrapedData(NamedTuple):
    page_number: int
    page_url: str
    group_id: int
    group_index: int
    element_index: int
    # actual collected data
    data: Dict


def scraped_data_sorter(data: ScrapedData) -> Tuple[int, int, int, int]:
    return data.page_number, data.group_index, data.group_id, data.element_index


def scraped_data_grouper(data: ScrapedData) -> Tuple[int, str, int, int, int]:
    return data.page_number, data.page_url, data.group_index, data.group_id, data.element_index



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dude', 'dude.optional']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'braveblock>=0.1.13,<0.4.0', 'playwright>=1.19.0,<2.0.0']

extras_require = \
{'bs4': ['beautifulsoup4>=4.10.0,<5.0.0', 'httpx>=0.22,<0.24'],
 'lxml': ['cssselect>=1.1.0,<2.0.0', 'httpx>=0.22,<0.24', 'lxml>=4.8.0,<5.0.0'],
 'parsel': ['httpx>=0.22,<0.24', 'parsel>=1.6.0,<2.0.0'],
 'pyppeteer': ['pyppeteer>=1.0.2,<2.0.0'],
 'selenium': ['pybrowsers>=0.4.1,<0.5.0',
              'selenium-wire>=4.6.2,<5.0.0',
              'webdriver-manager>=3.7.0,<4.0.0']}

entry_points = \
{'console_scripts': ['dude = dude:cli']}

setup_kwargs = {
    'name': 'pydude',
    'version': '0.22.0',
    'description': 'dude uncomplicated data extraction',
    'long_description': '<table>\n    <tr>\n        <td>License</td>\n        <td><img src=\'https://img.shields.io/pypi/l/pydude.svg?style=for-the-badge\' alt="License"></td>\n        <td>Version</td>\n        <td><img src=\'https://img.shields.io/pypi/v/pydude.svg?logo=pypi&style=for-the-badge\' alt="Version"></td>\n    </tr>\n    <tr>\n        <td>Github Actions</td>\n        <td><img src=\'https://img.shields.io/github/workflow/status/roniemartinez/dude/Python?label=actions&logo=github%20actions&style=for-the-badge\' alt="Github Actions"></td>\n        <td>Coverage</td>\n        <td><img src=\'https://img.shields.io/codecov/c/github/roniemartinez/dude/branch?label=codecov&logo=codecov&style=for-the-badge\' alt="CodeCov"></td>\n    </tr>\n    <tr>\n        <td>Supported versions</td>\n        <td><img src=\'https://img.shields.io/pypi/pyversions/pydude.svg?logo=python&style=for-the-badge\' alt="Python Versions"></td>\n        <td>Wheel</td>\n        <td><img src=\'https://img.shields.io/pypi/wheel/pydude.svg?style=for-the-badge\' alt="Wheel"></td>\n    </tr>\n    <tr>\n        <td>Status</td>\n        <td><img src=\'https://img.shields.io/pypi/status/pydude.svg?style=for-the-badge\' alt="Status"></td>\n        <td>Downloads</td>\n        <td><img src=\'https://img.shields.io/pypi/dm/pydude.svg?style=for-the-badge\' alt="Downloads"></td>\n    </tr>\n    <tr>\n        <td>All Contributors</td>\n        <td><a href="#contributors-"><img src=\'https://img.shields.io/github/all-contributors/roniemartinez/dude?style=for-the-badge\' alt="All Contributors"></a></td>\n    </tr>\n</table>\n\n# dude uncomplicated data extraction\n\nDude is a very simple framework for writing web scrapers using Python decorators.\nThe design, inspired by [Flask](https://github.com/pallets/flask), was to easily build a web scraper in just a few lines of code.\nDude has an easy-to-learn syntax.\n\n> 🚨 Dude is currently in Pre-Alpha. Please expect breaking changes.\n\n## Installation\n\nTo install, simply run the following from terminal.\n\n```bash\npip install pydude\nplaywright install  # Install playwright binaries for Chrome, Firefox and Webkit.\n```\n\n## Minimal web scraper\n\nThe simplest web scraper will look like this:\n\n```python\nfrom dude import select\n\n\n@select(css="a")\ndef get_link(element):\n    return {"url": element.get_attribute("href")}\n```\n\nThe example above will get all the [hyperlink](https://en.wikipedia.org/wiki/Hyperlink#HTML) elements in a page and calls the handler function `get_link()` for each element.\n\n## How to run the scraper\n\nYou can run your scraper from terminal/shell/command-line by supplying URLs, the output filename of your choice and the paths to your python scripts to `dude scrape` command.\n\n```bash\ndude scrape --url "<url>" --output data.json path/to/script.py\n```\n\nThe output in `data.json` should contain the actual URL and the metadata prepended with underscore.\n\n```json5\n[\n  {\n    "_page_number": 1,\n    "_page_url": "https://dude.ron.sh/",\n    "_group_id": 4502003824,\n    "_group_index": 0,\n    "_element_index": 0,\n    "url": "/url-1.html"\n  },\n  {\n    "_page_number": 1,\n    "_page_url": "https://dude.ron.sh/",\n    "_group_id": 4502003824,\n    "_group_index": 0,\n    "_element_index": 1,\n    "url": "/url-2.html"\n  },\n  {\n    "_page_number": 1,\n    "_page_url": "https://dude.ron.sh/",\n    "_group_id": 4502003824,\n    "_group_index": 0,\n    "_element_index": 2,\n    "url": "/url-3.html"\n  }\n]\n```\n\nChanging the output to `--output data.csv` should result in the following CSV content.\n\n![data.csv](docs/csv.png)\n\n## Features\n\n- Simple [Flask](https://github.com/pallets/flask)-inspired design - build a scraper with decorators.\n- Uses [Playwright](https://playwright.dev/python/) API - run your scraper in Chrome, Firefox and Webkit and leverage Playwright\'s powerful selector engine supporting CSS, XPath, text, regex, etc.\n- Data grouping - group related results.\n- URL pattern matching - run functions on matched URLs.\n- Priority - reorder functions based on priority.\n- Setup function - enable setup steps (clicking dialogs or login).\n- Navigate function - enable navigation steps to move to other pages.\n- Custom storage - option to save data to other formats or database.\n- Async support - write async handlers.\n- Option to use other parser backends aside from Playwright.\n  - [BeautifulSoup4](https://roniemartinez.github.io/dude/advanced/09_beautifulsoup4.html) - `pip install pydude[bs4]`\n  - [Parsel](https://roniemartinez.github.io/dude/advanced/10_parsel.html) - `pip install pydude[parsel]`\n  - [lxml](https://roniemartinez.github.io/dude/advanced/11_lxml.html) - `pip install pydude[lxml]`\n  - [Pyppeteer](https://roniemartinez.github.io/dude/advanced/12_pyppeteer.html) - `pip install pydude[pyppeteer]`\n  - [Selenium](https://roniemartinez.github.io/dude/advanced/13_selenium.html) - `pip install pydude[selenium]`\n- Option to follow all links indefinitely (Crawler/Spider).\n- Events - attach functions to startup, pre-setup, post-setup and shutdown events.\n- Option to save data on every page.\n\n## Supported Parser Backends\n\nBy default, Dude uses Playwright but gives you an option to use parser backends that you are familiar with.\nIt is possible to use parser backends like \n[BeautifulSoup4](https://roniemartinez.github.io/dude/advanced/09_beautifulsoup4.html), \n[Parsel](https://roniemartinez.github.io/dude/advanced/10_parsel.html),\n[lxml](https://roniemartinez.github.io/dude/advanced/11_lxml.html),\n[Pyppeteer](https://roniemartinez.github.io/dude/advanced/12_pyppeteer.html), \nand [Selenium](https://roniemartinez.github.io/dude/advanced/13_selenium.html).\n\nHere is the summary of features supported by each parser backend.\n\n<table>\n<thead>\n  <tr>\n    <td rowspan="2" style=\'text-align:center;\'>Parser Backend</td>\n    <td rowspan="2" style=\'text-align:center;\'>Supports<br>Sync?</td>\n    <td rowspan="2" style=\'text-align:center;\'>Supports<br>Async?</td>\n    <td colspan="4" style=\'text-align:center;\'>Selectors</td>\n    <td rowspan="2" style=\'text-align:center;\'><a href="https://roniemartinez.github.io/dude/advanced/01_setup.html">Setup<br>Handler</a></td>\n    <td rowspan="2" style=\'text-align:center;\'><a href="https://roniemartinez.github.io/dude/advanced/02_navigate.html">Navigate<br>Handler</a></td>\n  </tr>\n  <tr>\n    <td>CSS</td>\n    <td>XPath</td>\n    <td>Text</td>\n    <td>Regex</td>\n  </tr>\n</thead>\n<tbody>\n  <tr>\n    <td>Playwright</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n  </tr>\n  <tr>\n    <td>BeautifulSoup4</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>🚫</td>\n    <td>🚫</td>\n    <td>🚫</td>\n    <td>🚫</td>\n    <td>🚫</td>\n  </tr>\n  <tr>\n    <td>Parsel</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>🚫</td>\n    <td>🚫</td>\n  </tr>\n  <tr>\n    <td>lxml</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>🚫</td>\n    <td>🚫</td>\n  </tr>\n  <tr>\n    <td>Pyppeteer</td>\n    <td>🚫</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>🚫</td>\n    <td>✅</td>\n    <td>✅</td>\n  </tr>\n  <tr>\n    <td>Selenium</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>✅</td>\n    <td>🚫</td>\n    <td>✅</td>\n    <td>✅</td>\n  </tr>\n</tbody>\n</table>\n\n## Using the Docker image\n\nPull the docker image using the following command.\n\n```console\ndocker pull roniemartinez/dude\n```\n\nAssuming that `script.py` exist in the current directory, run Dude using the following command.\n\n```console\ndocker run -it --rm -v "$PWD":/code roniemartinez/dude dude scrape --url <url> script.py\n```\n\n## Documentation\n\nRead the complete documentation at [https://roniemartinez.github.io/dude/](https://roniemartinez.github.io/dude/).\nAll the advanced and useful features are documented there.\n\n## Requirements\n\n- ✅ Any dude should know how to work with selectors (CSS or XPath).\n- ✅ Familiarity with any backends that you love (see [Supported Parser Backends](#supported-parser-backends))\n- ✅ Python decorators... you\'ll live, dude!\n\n## Why name this project "dude"?\n\n- ✅ A [Recursive acronym](https://en.wikipedia.org/wiki/Recursive_acronym) looks nice.\n- ✅ Adding "uncomplicated" (like [`ufw`](https://wiki.ubuntu.com/UncomplicatedFirewall)) into the name says it is a very simple framework. \n- ✅ Puns! I also think that if you want to do web scraping, there\'s probably some random dude around the corner who can make it very easy for you to start with it. 😊\n\n## Author\n\n[Ronie Martinez](mailto:ronmarti18@gmail.com)\n\n## Contributors ✨\n\nThanks goes to these wonderful people ([emoji key](https://allcontributors.org/docs/en/emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://ron.sh"><img src="https://avatars.githubusercontent.com/u/2573537?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Ronie Martinez</b></sub></a><br /><a href="#maintenance-roniemartinez" title="Maintenance">🚧</a> <a href="https://github.com/roniemartinez/dude/commits?author=roniemartinez" title="Code">💻</a> <a href="https://github.com/roniemartinez/dude/commits?author=roniemartinez" title="Documentation">📖</a> <a href="#infra-roniemartinez" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind welcome!',
    'author': 'Ronie Martinez',
    'author_email': 'ronmarti18@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/roniemartinez/dude',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import time
from typing import Iterable, Optional, Tuple

import httpx
from httpx import Request

logger = logging.getLogger(__name__)


async def async_http_get(client: httpx.AsyncClient, request: Request) -> Tuple[Optional[str], str]:
    try:
        response = await client.send(request)
        response.raise_for_status()
        return response.text, str(response.url)
    except (httpx.HTTPStatusError, httpx.RequestError) as e:
        logger.warning(e)
        return None, str(request.url)


def http_get(client: httpx.Client, request: Request) -> Tuple[Optional[str], str]:
    try:
        response = client.send(request)
        response.raise_for_status()
        return response.text, str(response.url)
    except (httpx.HTTPStatusError, httpx.RequestError) as e:
        logger.warning(e)
        return None, str(request.url)


class HTTPXMixin:
    def _block_httpx_request_if_needed(self, request: Request) -> None:
        url = str(request.url)
        source_url = (
            request.headers.get("referer") or request.headers.get("origin") or request.headers.get("host") or url
        )
        if self.adblock.check_network_urls(  # type: ignore
            url=url,
            source_url=source_url,
            request_type=request.headers.get("sec-fetch-dest") or "other",
        ):
            logger.info("URL %s has been blocked.", url)
            raise httpx.RequestError(message=f"URL {url} has been blocked.", request=request)

    async def _async_block_httpx_request_if_needed(self, request: Request) -> None:
        self._block_httpx_request_if_needed(request)

    def iter_requests(self) -> Iterable[Request]:
        try:
            while True:
                try:
                    url = next(self.iter_urls())  # type: ignore
                    yield Request(method="GET", url=url)
                    continue
                except StopIteration:
                    pass
                request = self.requests.popleft()  # type: ignore
                can_fetch, crawl_delay = self.can_fetch_and_crawl_delay(str(request.url))  # type: ignore
                if not can_fetch:
                    logger.info("Not allowed to crawl %s", str(request.url))
                    continue
                time.sleep(crawl_delay)
                self.current_url = str(request.url)  # type: ignore
                yield request
        except IndexError:
            pass


def get_chromedriver_latest_release() -> str:
    """
    https://chromedriver.chromium.org/downloads/version-selection
    """
    import browsers

    version = browsers.get("chrome")["version"].rsplit(".", 1)[0]

    with httpx.Client() as client:
        try:
            response = client.get(f"https://chromedriver.storage.googleapis.com/LATEST_RELEASE_{version}")
            response.raise_for_status()
            latest = response.text.strip()
            logger.info("Matching ChromeDriver latest release is %s", latest)
            return latest
        except httpx.HTTPStatusError as e:
            logger.exception(e)

        return "latest"
